import os
from dataclasses import dataclass

import pandas as pd


@dataclass
class DiscoveryContent:
    """
    Container for content generated by the Discovery module.
    """

    pandas_general_description: str
    pandas_categorical_description: pd.DataFrame
    pandas_numerical_description: pd.DataFrame
    discovery: str = ""

    @property
    def pandas_response(self) -> str:
        """
        A Discovery run response containing only Pandas information.

        Returns
        -------
        str
            The response string.
        """
        return f"""{self.pandas_general_description}

{self.pandas_categorical_description}

{self.pandas_numerical_description}"""

    def __str__(self) -> str:
        return f"""
Data General Info
{self.pandas_general_description}

Numeric Data Descriptions
{self.pandas_numerical_description}

Categorical Data Descriptions
{self.pandas_categorical_description}

LLM Generated Discovery
{self.discovery}
            """

    def to_txt(
        self,
        file_dir: str = "./",
        file_name: str = "discovery.txt",
        include_pandas: bool = True,
    ) -> None:
        """
        Write the generated discovery content to a .txt file.

        Parameters
        ----------
        file_dir : str, optional
            The file directory to write to, by default "./"
        file_name : str, optional
            The name of the file, by default "discovery.txt"
        include_pandas : bool
            Whether to include the Pandas summaries, by default True
        """

        assert file_name.endswith(".txt"), "File name is not .txt"

        self._export_to_file(
            file_name=file_name, file_dir=file_dir, include_pandas=include_pandas
        )

    def to_markdown(
        self,
        file_dir: str = "./",
        file_name: str = "discovery.md",
        include_pandas: bool = True,
    ) -> None:
        """
        Write the generated discovery content to a Markdown file.

        Parameters
        ----------
        file_dir : str, optional
            The file directory to write to, by default "./"
        file_name : str, optional
            The name of the file, by default "discovery.md"
        include_pandas : bool
            Whether to include the Pandas summaries, by default True
        """

        assert file_name.endswith(".md"), "File name is not .md"

        self._export_to_file(
            file_name=file_name, file_dir=file_dir, include_pandas=include_pandas
        )

    def _export_to_file(
        self, file_name: str, file_dir: str = "./", include_pandas: bool = True
    ) -> None:
        if file_dir != "./":
            os.makedirs(file_dir, exist_ok=True)

        with open(f"./{file_dir}{file_name}", "w") as f:
            if include_pandas:
                f.write(self.__str__())
            else:
                f.write(self.discovery)
